==> elmo-2.0.1.tgz: package/src/server/config/webpack.tsx <==
//

/*
 * Module dependencies
 */

// import // development as babelDev,
// production as babelPro
// typescript as babelTS
// './babel'
import { Route } from '../../elmo'
import StaticSiteGeneratorPlugin from '../plugins/static'
// import Watcher from '../plugins/watcher'
// import WatchFix from '../plugins/watch-fix'
import * as CopyWebpackPlugin from 'copy-webpack-plugin'
import * as readPkg from 'read-pkg-up'
// import * as MessagesPlugin from 'webpack-messages'
import rewrite from '../../utils/rewrite-route'
import resolves from '../../utils/resolves'
import glob from '../../utils/glob'

import {
  join,
  resolve,
  basename,
  dirname,
  relative,
  extname,
  sep as PATH_SEP
} from 'path'
import { exists } from 'mz/fs'
// import Route from '../../lib/route'
import * as webpack from 'webpack'
// import Messages from '../plugins/messages'

/**
 * CWD
 */

const cwd = process.cwd()

/**
 * Environment Variables
 */

function Environment(): { [s: string]: string } {
  return Object.keys(process.env).reduce(function(
    o: { [s: string]: string },
    k: string
  ) {
    o['process.env.' + k] = JSON.stringify(process.env[k])
    return o
  },
  {})
}

/**
 * Aliases
 */

type Aliases = {
  [alias: string]: string
}

// NOTE: these are required server-side
// so the server will require them directly
// and know where they are.
// (e.g. require('elmo/router'))
function ServerAliases(projectRoot: string, elmoRoot: string): Aliases {
  // resolve the project's preact first
  // then look for elmo's preact if
  // we can't find it
  const preact = resolves(
    join(projectRoot, 'node_modules', 'preact'),
    join(elmoRoot, 'node_modules', 'preact')
  )
  if (!preact) {
    throw new Error('no preact...')
  }

  return {
    react: preact,
    'react-dom': preact,
    preact: preact,
    elmo: join(__dirname, '..', '..', 'elmo.js'),
    style: join(__dirname, '..', '..', 'lib', 'style', 'style.js'),
    tslib: join(elmoRoot, 'node_modules', 'tslib')
  }
}

// NOTE: these are required client-side
// (e.g. import Router from 'elmo/router')
// these should only be requires that are
// added to the build. the public-facing
// aliases are here to allow us to narrow
// down our elmo typescript loader
function BrowserAliases(projectRoot: string, elmoRoot: string): Aliases {
  // TODO: this should only work in development,
  // then have a helpful error when you go into
  // production.
  const preact = resolves(
    join(projectRoot, 'node_modules', 'preact'),
    join(elmoRoot, 'node_modules', 'preact')
  )
  // TODO: don't throw
  if (!preact) {
    throw new Error('no preact...')
  }

  return {
    preact: preact,
    react: preact,
    'react-dom': preact,
    unfetch: join(elmoRoot, 'node_modules', 'unfetch'),
    'promise-polyfill': join(elmoRoot, 'node_modules', 'promise-polyfill'),
    elmo: join(__dirname, '..', '..', 'elmo.js'),
    style: join(__dirname, '..', '..', 'lib', 'style', 'style.js'),
    tslib: join(elmoRoot, 'node_modules', 'tslib'),

    // TODO: clean this up:
    // these requirements are for the hot-middleware-client
    // and should only be pulled in during development
    // and when the action is serve
    'webpack-hot-middleware': join(
      elmoRoot,
      'node_modules',
      'webpack-hot-middleware'
    ),
    'ansi-html': join(elmoRoot, 'node_modules', 'ansi-html'),
    'strip-ansi': join(elmoRoot, 'node_modules', 'strip-ansi'),
    'ansi-regex': join(elmoRoot, 'node_modules', 'ansi-regex'),
    'html-entities': join(elmoRoot, 'node_modules', 'html-entities')
  }
}

/**
 * Input params
 */

export namespace Webpack {
  export interface Input {
    aliases: { [module: string]: string }
    environment: 'production' | 'development'
    routes: Route[]
    target: string
    action: string
    root: string
  }
}

/**
 * Webpack Config
 */

export default async function Webpack(
  input: Webpack.Input
): Promise<Error | webpack.Configuration> {
  const { path: pkgPath } = await readPkg({ cwd: __dirname })
  const ELMO_ROOT = dirname(pkgPath)

  const { target, environment, action, routes, aliases } = input
  const root = resolve(process.cwd(), input.root || '.')

  const serverAliases = Object.assign(ServerAliases(root, ELMO_ROOT), aliases)
  const browserAliases = Object.assign(BrowserAliases(root, ELMO_ROOT), aliases)

  const config: webpack.Configuration = {
    mode: environment,
    name: target,
    context: root,
    module: { rules: [] },
    plugins: [],
    entry: {}
  }

  const pages = await glob(root, join('pages', '**'))

  // filter out any page that contains an underscore in
  // any part of their path
  const publicPages = pages.filter((page: string) => {
    const rel = relative(join(root, 'pages'), page)
    const parts = rel.split(PATH_SEP)
    for (let i = 0; i < parts.length; i++) {
      const part = parts[i]
      if (part[0] === '_') return false
    }
    return true
  })

  if (publicPages.length === 0) {
    const rel = relative(cwd, root)
    return new Error(
      `no public pages found after searching: './${rel}/pages/**/*.{js,jsx,ts,tsx}'. Are you running from an elmo directory?`
    )
  }

  // not used right now, but perhaps useful in the future
  const common: string[] = []

  // setup the entries
  config.entry = publicPages.reduce(
    (obj: { [s: string]: string[] }, page: string) => {
      const name = rewrite(root, target, page)
      const bundle: string[] = []

      // add in the common bundles
      bundle.push(...common)

      // only include webpack hot middleware when we're
      // not in production and using the dev server
      // and we're building for the browser
      if (
        action === 'serve' &&
        environment !== 'production' &&
        target === 'browser'
      ) {
        bundle.push(
          'webpack-hot-middleware/client?path=/__webpack_hmr&noInfo=true&name=' +
            target
        )
      }

      obj[name] = bundle.concat(page)
      return obj
    },
    {}
  )

  // add in the document page if we're on the server
  if (target === 'server') {
    const document = pages.find(p => basename(p, extname(p)) === '_document')
    config.entry['_document.html'] =
      document || join(__dirname, '../../lib/_document.js')
  }

  config.target = target === 'browser' ? 'web' : 'node'

  if (target !== 'browser') {
    // this is to ensure that we use node's actual require(...)
    // which is needed for styled-jsx, but should probably be the
    // default anyway.
    config.externals = [
      function(_parent, request, fn) {
        if (serverAliases[request]) {
          return fn(null, serverAliases[request])
        } else if (request[0] !== '.' && request[0] !== '/') {
          // externalize all node_modules
          return fn(null, request[0])
        } else {
          // pass relative paths through (don't externalize)
          return fn(null, undefined)
        }
      }
    ]
  }

  if (target === 'browser') {
    config.node = {
      console: false,
      global: false,
      process: false,
      __filename: false,
      __dirname: false,
      Buffer: false,
      setImmediate: false,
      dns: false,
      fs: false,
      path: true,
      url: false
    }

    // support electron renderer and
    // web support seamlessly
    config.externals = [
      function(_parent, request, fn) {
        if (request === 'electron') {
          return fn(null, request)
        }
        return fn(null, undefined)
      }
    ]
  }

  // resolves
  config.resolve = {
    extensions: ['.ts', '.js', '.tsx', '.jsx'],
    modules: [join(root, 'node_modules')],
    alias: browserAliases
  }

  // where to resolve loaders
  config.resolveLoader = {
    modules: [join(root, 'node_modules')]
  }

  /**
   * Output
   */

  config.output = {
    path: '/',
    publicPath: '/',
    // [name] ensures the path matches the entry key
    filename: '[name]',
    libraryTarget: 'umd'
  }

  /**
   * Rules
   */

  let rules = (config.module as webpack.Module).rules

  // support json files
  rules.push({
    test: /\.json$/,
    use: ['json-loader']
  })

  // support loading .jsx files
  rules.push({
    test: /\.jsx/,
    exclude: /(node_modules|bower_components)/,
    loader: join(ELMO_ROOT, 'node_modules', 'babel-loader'),
    options: {
      plugins: [
        // TODO: check that these plugins exist
        join(
          ELMO_ROOT,
          'node_modules',
          'babel-plugin-transform-class-properties'
        ),
        [
          join(ELMO_ROOT, 'node_modules', 'babel-plugin-transform-react-jsx'),
          { pragma: 'h' }
        ]
      ]
    }
  })

  // automatically add import { h } from 'preact' to jsx pages
  // NOTE: we add this to tsx pages, but typescript will
  // complain about not being able to resolve anyway, but
  // it makes for simpler tests to include it
  rules.push({
    test: /\.[jt]sx/,
    exclude: /(node_modules|bower_components)/,
    loader: join(__dirname, '..', 'loaders', 'h.js')
  })

  // compile our project's typescript files
  rules.push({
    test: /\.tsx?$/,
    exclude: /(node_modules|bower_components)/,
    loader: join(__dirname, '..', 'loaders', 'typescript.js'),
    query: { environment, target, root }
  })

  // compile .mdx files
  // rules.push({
  //   test: /\.mdx$/,
  //   include: publicPages.map(page => resolve(root, page)),
  //   loader: join(ELMO_JS_SRC, 'server', 'loaders', 'mdx.js'),
  //   query: { root }
  // })
  // if (target === 'browser') {
  //   rules.push({
  //     test: /\.mdx$/,
  //     include: publicPages.map(page => resolve(root, page)),
  //     loader: join(ELMO_JS_SRC, 'server', 'loaders', 'boot.js'),
  //     query: { root }
  //   })
  // }

  // certain typescript files are allowed to be compiled
  // by the frontend configuration. allow certain
  // modules to be compiled from elmo's ts source
  // rules.push({
  //   test: /\.tsx?$/,
  //   include: [
  //     join(ELMO_TS_SRC, 'elmo.ts'),
  //     join(ELMO_TS_SRC, 'lib'),
  //     join(ELMO_TS_SRC, 'client')
  //   ],
  //   loader: typescript_loader,
  //   query: { environment, target, root }
  // })

  if (target === 'browser') {
    // always polyfill to find bugs faster
    // right now just polyfilling fetch and promise
    rules.push({
      test: /\.[tj]sx?$/,
      include: publicPages.map((page: string) => resolve(root, page)),
      loader: join(__dirname, '..', 'loaders', 'polyfills.js')
    })

    // boot script for client-side, only added to the top-level pages
    rules.push({
      test: /\.[jt]sx?$/,
      include: publicPages.map(page => resolve(root, page)),
      use: {
        loader: join(__dirname, '..', 'loaders', 'boot.js'),
        query: { root, routes }
      }
    })
  }

  /**
   * Reverse plugins, webpack goes from bottom to top
   * but top to bottom makes more sense in a config
   */

  rules = rules.reverse()

  /**
   * Plugins
   */

  // plugins
  const plugins = config.plugins || []

  // get all the environment variables
  let constants = Environment()

  // attach the environment to the NODE_ENV
  constants['process.env.NODE_ENV'] = JSON.stringify(
    constants['process.env.NODE_ENV'] || environment
  )

  // // add the routes into the router
  // if (target === 'browser') {
  //   constants = Object.assign(constants, {
  //     _ELMO_ROUTES_: JSON.stringify(routes.length ? routes : await Routes(root))
  //   })
  // }

  // Define environment variables
  plugins.push(new webpack.DefinePlugin(constants))

  // Don't write files if there are errors
  plugins.push(new webpack.NoEmitOnErrorsPlugin())

  // use names instead of IDs
  plugins.push(new webpack.NamedModulesPlugin())

  // fix watching

  // server-side
  if (target === 'server') {
    plugins.push(
      new StaticSiteGeneratorPlugin({
        elmoRoot: ELMO_ROOT,
        routes,
        environment,
        target,
        root,
        action
      })
    )
  }

  // setup minification & common chunks
  if (target === 'browser') {
    ;(config as any).optimization = {
      minimize: environment === 'production',
      splitChunks: {
        name: 'common',
        filename: 'common.js',
        chunks: 'initial',

        // TODO: these options should be optimized
        // right now they make sure we always have
        // a common.js, but sometimes it's not
        // necessary.
        minChunks: 1,
        minSize: 0
      }
    }
  }

  // copy static assets over when we're building (doesn't matter what target we use)
  if (target === 'server') {
    // CopyWebpackPlugin will produce an error if this directory doesn't exist already
    // so we check it first
    if (await exists(join(root, 'static'))) {
      plugins.push(
        new CopyWebpackPlugin([
          {
            context: root,
            from: './static',
            to: './static',
            ignore: ['.*']
          }
        ])
      )
    }

    // copy over the _redirects
    if (await exists(join(root, '_redirects'))) {
      plugins.push(
        new CopyWebpackPlugin(
          [
            {
              context: root,
              from: '_redirects',
              to: '_redirects',
              toType: 'file'
            }
          ],
          { copyUnmodified: true }
        )
      )
    }
  }

  if (action === 'serve' && target === 'browser') {
    // when serving, show friendly errors
    // plugins.push(new MessagesPlugin())
  }

  // enable hot module replacement when in dev server in the browser
  if (
    action === 'serve' &&
    environment !== 'production' &&
    target === 'browser'
  ) {
    plugins.push(new webpack.HotModuleReplacementPlugin())
  }

  // minify scripts in production in the browser
  if (environment === 'production' && target === 'browser') {
    plugins.push(
      new webpack.LoaderOptionsPlugin({
        minimize: true,
        debug: false
      })
    )
  }

  // config.devtool = environment === 'development' ? 'cheap-source-map' : false
  config.devtool = false

  return config
}

==> elmo-2.0.2.tgz: package/src/server/config/webpack.tsx <==
//

/*
 * Module dependencies
 */

// import // development as babelDev,
// production as babelPro
// typescript as babelTS
// './babel'
import { Route } from '../../elmo'
import StaticSiteGeneratorPlugin from '../plugins/static'
// import Watcher from '../plugins/watcher'
// import WatchFix from '../plugins/watch-fix'
import * as CopyWebpackPlugin from 'copy-webpack-plugin'
import * as readPkg from 'read-pkg-up'
// import * as MessagesPlugin from 'webpack-messages'
import rewrite from '../../utils/rewrite-route'
import glob from '../../utils/glob'

import {
  join,
  resolve,
  basename,
  dirname,
  relative,
  extname,
  sep as PATH_SEP
} from 'path'
import { exists } from 'mz/fs'
// import Route from '../../lib/route'
import * as webpack from 'webpack'
// import Messages from '../plugins/messages'

/**
 * CWD
 */

const cwd = process.cwd()

/**
 * Environment Variables
 */

function Environment(): { [s: string]: string } {
  return Object.keys(process.env).reduce(function(
    o: { [s: string]: string },
    k: string
  ) {
    o['process.env.' + k] = JSON.stringify(process.env[k])
    return o
  },
  {})
}

/**
 * Aliases
 */

type Aliases = {
  [alias: string]: string
}

// NOTE: these are required server-side
// so the server will require them directly
// and know where they are.
// (e.g. require('elmo/router'))
function ServerAliases(_projectRoot: string, _elmoRoot: string): Aliases {
  return {
    react: 'preact',
    'react-dom': 'preact',
    elmo: join(__dirname, '..', '..', 'elmo.js'),
    style: join(__dirname, '..', '..', 'lib', 'style', 'style.js')
  }
}

// NOTE: these are required client-side
// (e.g. import Router from 'elmo/router')
// these should only be requires that are
// added to the build. the public-facing
// aliases are here to allow us to narrow
// down our elmo typescript loader
function BrowserAliases(_projectRoot: string, _elmoRoot: string): Aliases {
  return {
    react: 'preact',
    'react-dom': 'preact',
    elmo: join(__dirname, '..', '..', 'elmo.js'),
    style: join(__dirname, '..', '..', 'lib', 'style', 'style.js')
  }
}

/**
 * Input params
 */

export namespace Webpack {
  export interface Input {
    aliases: { [module: string]: string }
    environment: 'production' | 'development'
    routes: Route[]
    target: string
    action: string
    root: string
  }
}

/**
 * Webpack Config
 */

export default async function Webpack(
  input: Webpack.Input
): Promise<Error | webpack.Configuration> {
  const { path: pkgPath } = await readPkg({ cwd: __dirname })
  const ELMO_ROOT = dirname(pkgPath)

  const { target, environment, action, routes, aliases } = input
  const root = resolve(process.cwd(), input.root || '.')

  const serverAliases = Object.assign(ServerAliases(root, ELMO_ROOT), aliases)
  const browserAliases = Object.assign(BrowserAliases(root, ELMO_ROOT), aliases)

  const config: webpack.Configuration = {
    mode: environment,
    name: target,
    context: root,
    module: { rules: [] },
    plugins: [],
    entry: {}
  }

  const pages = await glob(root, join('pages', '**'))

  // filter out any page that contains an underscore in
  // any part of their path
  const publicPages = pages.filter((page: string) => {
    const rel = relative(join(root, 'pages'), page)
    const parts = rel.split(PATH_SEP)
    for (let i = 0; i < parts.length; i++) {
      const part = parts[i]
      if (part[0] === '_') return false
    }
    return true
  })

  if (publicPages.length === 0) {
    const rel = relative(cwd, root)
    return new Error(
      `no public pages found after searching: './${rel}/pages/**/*.{js,jsx,ts,tsx}'. Are you running from an elmo directory?`
    )
  }

  // not used right now, but perhaps useful in the future
  const common: string[] = []

  // setup the entries
  config.entry = publicPages.reduce(
    (obj: { [s: string]: string[] }, page: string) => {
      const name = rewrite(root, target, page)
      const bundle: string[] = []

      // add in the common bundles
      bundle.push(...common)

      // only include webpack hot middleware when we're
      // not in production and using the dev server
      // and we're building for the browser
      if (
        action === 'serve' &&
        environment !== 'production' &&
        target === 'browser'
      ) {
        bundle.push(
          'webpack-hot-middleware/client?path=/__webpack_hmr&noInfo=true&name=' +
            target
        )
      }

      obj[name] = bundle.concat(page)
      return obj
    },
    {}
  )

  // add in the document page if we're on the server
  if (target === 'server') {
    const document = pages.find(p => basename(p, extname(p)) === '_document')
    config.entry['_document.html'] =
      document || join(__dirname, '../../lib/_document.js')
  }

  config.target = target === 'browser' ? 'web' : 'node'

  if (target !== 'browser') {
    // this is to ensure that we use node's actual require(...)
    // which is needed for styled-jsx, but should probably be the
    // default anyway.
    config.externals = [
      function(_parent, request, fn) {
        if (serverAliases[request]) {
          return fn(null, serverAliases[request])
        } else if (request[0] !== '.' && request[0] !== '/') {
          // externalize all node_modules
          return fn(null, request)
        } else {
          // pass relative paths through (don't externalize)
          return fn(null, undefined)
        }
      }
    ]
  }

  if (target === 'browser') {
    config.node = {
      console: false,
      global: false,
      process: false,
      __filename: false,
      __dirname: false,
      Buffer: false,
      setImmediate: false,
      dns: false,
      fs: false,
      path: true,
      url: false
    }

    // support electron renderer and
    // web support seamlessly
    config.externals = [
      function(_parent, request, fn) {
        if (request === 'electron') {
          return fn(null, request)
        }
        return fn(null, undefined)
      }
    ]
  }

  // resolves
  config.resolve = {
    extensions: ['.ts', '.js', '.tsx', '.jsx'],
    modules: [join(root, 'node_modules')],
    alias: browserAliases
  }

  // where to resolve loaders
  config.resolveLoader = {
    modules: [join(root, 'node_modules')]
  }

  /**
   * Output
   */

  config.output = {
    path: '/',
    publicPath: '/',
    // [name] ensures the path matches the entry key
    filename: '[name]',
    libraryTarget: 'umd'
  }

  /**
   * Rules
   */

  let rules = (config.module as webpack.Module).rules

  // support json files
  rules.push({
    test: /\.json$/,
    use: ['json-loader']
  })

  // support loading .jsx files
  rules.push({
    test: /\.jsx/,
    exclude: /(node_modules|bower_components)/,
    loader: join(ELMO_ROOT, 'node_modules', 'babel-loader'),
    options: {
      plugins: [
        // TODO: check that these plugins exist
        join(
          ELMO_ROOT,
          'node_modules',
          'babel-plugin-transform-class-properties'
        ),
        [
          join(ELMO_ROOT, 'node_modules', 'babel-plugin-transform-react-jsx'),
          { pragma: 'h' }
        ]
      ]
    }
  })

  // automatically add import { h } from 'preact' to jsx pages
  // NOTE: we add this to tsx pages, but typescript will
  // complain about not being able to resolve anyway, but
  // it makes for simpler tests to include it
  rules.push({
    test: /\.[jt]sx/,
    exclude: /(node_modules|bower_components)/,
    loader: join(__dirname, '..', 'loaders', 'h.js')
  })

  // compile our project's typescript files
  rules.push({
    test: /\.tsx?$/,
    exclude: /(node_modules|bower_components)/,
    loader: join(__dirname, '..', 'loaders', 'typescript.js'),
    query: { environment, target, root }
  })

  // compile .mdx files
  // rules.push({
  //   test: /\.mdx$/,
  //   include: publicPages.map(page => resolve(root, page)),
  //   loader: join(ELMO_JS_SRC, 'server', 'loaders', 'mdx.js'),
  //   query: { root }
  // })
  // if (target === 'browser') {
  //   rules.push({
  //     test: /\.mdx$/,
  //     include: publicPages.map(page => resolve(root, page)),
  //     loader: join(ELMO_JS_SRC, 'server', 'loaders', 'boot.js'),
  //     query: { root }
  //   })
  // }

  // certain typescript files are allowed to be compiled
  // by the frontend configuration. allow certain
  // modules to be compiled from elmo's ts source
  // rules.push({
  //   test: /\.tsx?$/,
  //   include: [
  //     join(ELMO_TS_SRC, 'elmo.ts'),
  //     join(ELMO_TS_SRC, 'lib'),
  //     join(ELMO_TS_SRC, 'client')
  //   ],
  //   loader: typescript_loader,
  //   query: { environment, target, root }
  // })

  if (target === 'browser') {
    // always polyfill to find bugs faster
    // right now just polyfilling fetch and promise
    rules.push({
      test: /\.[tj]sx?$/,
      include: publicPages.map((page: string) => resolve(root, page)),
      loader: join(__dirname, '..', 'loaders', 'polyfills.js')
    })

    // boot script for client-side, only added to the top-level pages
    rules.push({
      test: /\.[jt]sx?$/,
      include: publicPages.map(page => resolve(root, page)),
      use: {
        loader: join(__dirname, '..', 'loaders', 'boot.js'),
        query: { root, routes }
      }
    })
  }

  /**
   * Reverse plugins, webpack goes from bottom to top
   * but top to bottom makes more sense in a config
   */

  rules = rules.reverse()

  /**
   * Plugins
   */

  // plugins
  const plugins = config.plugins || []

  // get all the environment variables
  let constants = Environment()

  // attach the environment to the NODE_ENV
  constants['process.env.NODE_ENV'] = JSON.stringify(
    constants['process.env.NODE_ENV'] || environment
  )

  // // add the routes into the router
  // if (target === 'browser') {
  //   constants = Object.assign(constants, {
  //     _ELMO_ROUTES_: JSON.stringify(routes.length ? routes : await Routes(root))
  //   })
  // }

  // Define environment variables
  plugins.push(new webpack.DefinePlugin(constants))

  // Don't write files if there are errors
  plugins.push(new webpack.NoEmitOnErrorsPlugin())

  // use names instead of IDs
  plugins.push(new webpack.NamedModulesPlugin())

  // fix watching

  // server-side
  if (target === 'server') {
    plugins.push(
      new StaticSiteGeneratorPlugin({
        elmoRoot: ELMO_ROOT,
        routes,
        environment,
        target,
        root,
        action
      })
    )
  }

  // setup minification & common chunks
  if (target === 'browser') {
    ;(config as any).optimization = {
      minimize: environment === 'production',
      splitChunks: {
        name: 'common',
        filename: 'common.js',
        chunks: 'initial',

        // TODO: these options should be optimized
        // right now they make sure we always have
        // a common.js, but sometimes it's not
        // necessary.
        minChunks: 1,
        minSize: 0
      }
    }
  }

  // copy static assets over when we're building (doesn't matter what target we use)
  if (target === 'server') {
    // CopyWebpackPlugin will produce an error if this directory doesn't exist already
    // so we check it first
    if (await exists(join(root, 'static'))) {
      plugins.push(
        new CopyWebpackPlugin([
          {
            context: root,
            from: './static',
            to: './static',
            ignore: ['.*']
          }
        ])
      )
    }

    // copy over the _redirects
    if (await exists(join(root, '_redirects'))) {
      plugins.push(
        new CopyWebpackPlugin(
          [
            {
              context: root,
              from: '_redirects',
              to: '_redirects',
              toType: 'file'
            }
          ],
          { copyUnmodified: true }
        )
      )
    }
  }

  if (action === 'serve' && target === 'browser') {
    // when serving, show friendly errors
    // plugins.push(new MessagesPlugin())
  }

  // enable hot module replacement when in dev server in the browser
  if (
    action === 'serve' &&
    environment !== 'production' &&
    target === 'browser'
  ) {
    plugins.push(new webpack.HotModuleReplacementPlugin())
  }

  // minify scripts in production in the browser
  if (environment === 'production' && target === 'browser') {
    plugins.push(
      new webpack.LoaderOptionsPlugin({
        minimize: true,
        debug: false
      })
    )
  }

  // config.devtool = environment === 'development' ? 'cheap-source-map' : false
  config.devtool = false

  return config
}
